from dataclasses import dataclass, field

from abacus.core import AbacusError, AccountName, Amount, Credit, Debit, Posting, double

Numeric = int | float | Amount


@dataclass  # make class serialisable for EntryStore
class Entry:
    """An Entry class is a user interface for creatting and manipulating a multiple entry."""

    title: str
    posting: Posting = field(default_factory=list)
    is_closing: bool = False
    amount: Numeric | None = None

    def set_amount(self, amount: Numeric):
        """Set amount for the entry."""
        self.amount = Amount(amount)
        return self

    def get_amount(self, amount: Numeric | None = None) -> Amount:
        """Use provided amount, default amount or raise error if no data about amount."""
        if amount is None:
            if self.amount:
                return Amount(self.amount)
            else:
                raise AbacusError("Amount is not set.")
        return Amount(amount)

    def debit(self, account_name: AccountName, amount: Numeric | None = None):
        """Add debit part to entry."""
        self.posting.append(Debit(account_name, self.get_amount(amount)))
        return self

    def credit(self, account_name: AccountName, amount: Numeric | None = None):
        """Add credit part to entry."""
        self.posting.append(Credit(account_name, self.get_amount(amount)))
        return self

    def double(self, debit: AccountName, credit: AccountName, amount: Numeric):
        """Create double entry."""
        self.posting.extend(double(debit, credit, Amount(amount)))
        return self



from pathlib import Path

from pydantic import BaseModel, ConfigDict

from abacus.core import T5, AbacusError, AccountName, ChartDict, Pair


class SaveLoadMixin:
    """Class for loading and saving pydantic models to files."""

    @classmethod
    def load(cls, filename: str | Path):
        return cls.model_validate_json(Path(filename).read_text())  # type: ignore

    def save(self, filename: str | Path, allow_overwrite: bool = False):
        if not allow_overwrite and Path(filename).exists():
            raise FileExistsError(f"File already exists: {filename}")
        Path(filename).write_text(self.model_dump_json(indent=2))  # type: ignore


class Chart(BaseModel, SaveLoadMixin):
    """Chart of accounts."""

    model_config = ConfigDict(extra="forbid")

    retained_earnings: str
    current_earnings: str
    assets: list[str] = []
    capital: list[str] = []
    liabilities: list[str] = []
    income: list[str] = []
    expenses: list[str] = []
    contra_accounts: dict[str, list[str]] = {}
    names: dict[str, str] = {}

    @classmethod
    def default(cls):
        return cls(
            retained_earnings="retained_earnings", current_earnings="current_earnings"
        )

    def __post_init__(self):
        self.assert_all_account_names_are_unique()
        self.assert_no_invalid_contra_account_references()

    @property
    def regular_accounts(self):
        return (
            self.assets + self.capital + self.liabilities + self.income + self.expenses
        ) + [self.retained_earnings, self.current_earnings]

    @property
    def accounts(self):
        """All accounts in this chart including the duplicates."""
        return self.regular_accounts + sum(self.contra_accounts.values(), [])

    @property
    def duplicates(self):
        """Duplicate account names. Must be empty for valid chart."""
        names = self.accounts
        for name in set(names):
            names.remove(name)
        return names

    def assert_all_account_names_are_unique(self):
        """Raise error if any duplicate account names are found."""
        if names := self.duplicates:
            raise AbacusError(f"Account names are not unique: {names}")

    def assert_no_invalid_contra_account_references(self):
        """Verify chart by making an empty ledger and try closing it."""
        regular_account_names = self.regular_accounts
        for account_name in self.contra_accounts.keys():
            if account_name not in regular_account_names:
                raise AbacusError(
                    f"Account name should exist before making a contra account: {account_name}"
                )

    def _regular_accounts(self):
        for t, attr in (
            (T5.Asset, "assets"),
            (T5.Liability, "liabilities"),
            (T5.Capital, "capital"),
            (T5.Income, "income"),
            (T5.Expense, "expenses"),
        ):
            for account_name in getattr(self, attr):
                yield t, account_name

    def _contra_accounts(self):
        for account_name, contra_names in self.contra_accounts.items():
            for contra_name in contra_names:
                yield account_name, contra_name

    @property
    def mapping(self) -> ChartDict:
        """Create chart dictionary with unique account names."""
        chart_dict = ChartDict()
        for t, account_name in self._regular_accounts():
            chart_dict.set(t, account_name)
        chart_dict.set(T5.Capital, self.retained_earnings)
        chart_dict.set(T5.Capital, self.current_earnings)
        for account_name, contra_name in self._contra_accounts():
            chart_dict.offset(account_name, contra_name)
        return chart_dict

    def offset(self, account_name: str, contra_name: str):
        """Add contra account to chart."""
        self.contra_accounts.setdefault(account_name, list()).append(contra_name)
        return self

    def name(self, account_name: str, title: str):
        """Add descriptive account title."""
        self.names[account_name] = title
        return self

    def make_closing_pairs(self, accumulation_account: AccountName) -> list[Pair]:
        """Return list of tuples that allows to close ledger."""
        return list(self.mapping.closing_pairs(accumulation_account))

    # @property
    # def closing_pairs(self) -> list["Pair"]:
    #     """Return list of tuples that allows to close ledger at period end."""
    #     return self.make_closing_pairs(self.retained_earnings)



import json
from collections import UserDict
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Mapping, Sequence

from pydantic import BaseModel

from abacus.chart import Chart, SaveLoadMixin
from abacus.core import AbacusError, Amount
from abacus.entry import Entry


class BalancesDict(UserDict[str, Amount], SaveLoadMixin):
    """Dictionary of account names and balances.
    Mimics a pydantic model so that SaveLoadMixin can apply."""

    def model_dump_json(self, indent=2) -> str:
        return json.dumps(self.data, default=str, indent=indent)

    @staticmethod
    def coerce(d: Mapping[str, int | float | str | Amount]):
        return {k: Amount(v) for k, v in d.items()}

    @classmethod
    def model_validate_json(cls, s: str) -> "BalancesDict":
        return cls(cls.coerce(json.loads(s)))


class EntryStore(BaseModel, SaveLoadMixin):
    entries: list[Entry] = []


@dataclass
class PathFinder:
    """Deault paths for chart, store and balances files in a directory."""

    directory: str
    _chart: str = "chart.json"
    _store: str = "store.json"
    _balances: str = "balances.json"

    @property
    def base(self) -> Path:
        return Path(self.directory)

    @property
    def chart(self) -> Path:
        return self.base / self._chart

    @property
    def store(self) -> Path:
        return self.base / self._store

    @property
    def balances(self) -> Path:
        return self.base / self._balances

    def get_chart(self):
        try:
            return Chart.load(self.chart)
        except FileNotFoundError:
            raise AbacusError(f"Chart file not found: {self.chart}")

    def get_store(self):
        return EntryStore.load(self.store)

    def get_balances(self) -> BalancesDict:
        return BalancesDict.load(self.balances)


@dataclass
class Book:
    chart: Chart
    store: EntryStore = field(default_factory=EntryStore)

    def __post_init__(self):
        self.ledger = self.chart.mapping.to_ledger()
        self.opening_balances = {}
        self._income_statement = None

    def open(
        self,
        opening_balances: Mapping[str, str | int | float | Amount],
        opening_entry_title: str = "Opening entry",
    ):
        self.opening_balances = BalancesDict.coerce(opening_balances)
        entry = Entry(
            title=opening_entry_title,
            posting=self.chart.mapping.opening_entry(self.opening_balances),
        )
        self.post(entry)
        return self

    def post(self, entry: Entry):
        self.ledger.post(entry.posting)
        self.store.entries.append(entry)

    def post_many(self, entries: Sequence[Entry]):
        for entry in entries:
            self.post(entry)

    def is_closed(self):
        return self.ledger.is_closed(chart_dict=self.chart.mapping)

    def close(self, closing_entry_title: str = "Closing entry"):
        """Close ledger at period end."""
        # Persist income statement before income and expense accounts are deleted
        self._income_statement = self.income_statement
        # Make closing pairs
        closing_pairs = self.chart.make_closing_pairs(self.chart.current_earnings)
        last_pair = (self.chart.current_earnings, self.chart.retained_earnings)
        closing_pairs.append(last_pair)
        # Post closing entries to ledger
        entries = [
            Entry(closing_entry_title, posting=me, is_closing=True)
            for me in self.ledger.close(closing_pairs)
        ]
        # Store closing entries that were already posted to ledger
        self.store.entries.extend(entries)

    @property
    def income_statement(self):
        if self.is_closed():
            return self._income_statement
        return self.ledger.income_statement(self.chart.mapping)

    @property
    def balance_sheet(self):
        if self.is_closed():
            return self.ledger.balance_sheet(self.chart.mapping)
        else:
            # Close ledger up to current_earnings account
            temp_ledger = deepcopy(self.ledger)
            closing_pairs = self.chart.make_closing_pairs(self.chart.current_earnings)
            temp_ledger.close(closing_pairs)
            return temp_ledger.balance_sheet(self.chart.mapping)

    @property
    def trial_balance(self):
        return self.ledger.trial_balance

    @property
    def balances(self):
        return BalancesDict(**self.ledger.balances)

    @classmethod
    def load(cls, directory: str):
        """Load chart and starting balances from the directory."""
        path = PathFinder(directory)
        chart = path.get_chart()
        book = cls(chart)
        if path.balances.exists():
            opening_balances = path.get_balances()
            book.open(opening_balances)
        return book

    def save(self, directory: str, allow_overwrite: bool = False):
        """Save entries and period end balances to the directory."""
        path = PathFinder(directory)
        self.store.save(path.store, allow_overwrite)
        self.balances.save(path.balances, allow_overwrite)



"""Core accounting objects.

This module contains classes for:

  - chart of accounts (ChartDict),
  - general ledger (Ledger),
  - accounting entry (Posting),
  - summaries (TrialBalance) and financial reports (IncomeStatement, BalanceSheet).

Accounting workflow:

1. create chart of accounts and set current and retained earnings accounts
2. create ledger from chart with opening balances
3. post entries to ledger
4. show trial balance at any time
5. show proxy income statement and balance sheet if necessary
6. close ledger at accounting period end and make income statement
7. make post-close entries and make balance sheet
8. save permanent account balances for next period

Accounting conventions:

- regular accounts of five types (asset, liability, capital, income, expense),
- contra accounts to regular accounts are possible (eg depreciation, discounts, etc.).

Assumptions and simplifications (some may be relaxed in future versions):

- one currency
- one reporting period
- one level of accounts, no sub-accounts
- account names must be globally unique (eg cannot have two accounts named "other")
- chart always has current account and retained earnings account
- no account durations (current vs non-current)
- other comprehensive income account (OCIA) not calculated
- no journals, entries are posted to ledger directly
- an entry can touch any accounts
- entry amount can be positive, negative or zero
- net earnings are income less expenses, no gross profit or earnings before tax calculated
- period end closing will transfer current earnings to retained earnings
- no cash flow statement
- no statement of changes in equity
- no date or any transaction metadata recorded
"""

import decimal
from abc import ABC, abstractmethod
from collections import UserDict
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Iterator, Sequence, Type


class AbacusError(Exception):
    """Custom error for the abacus project."""


AccountName = str
Amount = decimal.Decimal
Pair = tuple[
    AccountName, AccountName
]  # ok to use tuple here, less useful as a dataclass


class T5(Enum):
    """Five types of accounts."""

    Asset = "asset"
    Liability = "liability"
    Capital = "capital"
    Income = "income"
    Expense = "expense"


@dataclass
class Regular:
    """Regular account, holds information about account type."""

    t: T5


@dataclass
class Contra:
    """Contra account, refers to an existing regular account."""

    name: str


class NotInChartError(AbacusError):
    pass


class NotInLedgerError(AbacusError):
    pass


class ChartDict(UserDict[str, Regular | Contra]):
    """Dictionary of accounts with their definitions.

    This is an intermediate data structure between Chart and Ledger
    that ensures account names are unique and contra accounts point to
    existing regular accounts.
    """

    # suggestion: may shut down __setitem__ for this class, and may call ChartMap

    def is_debit_account(self, account_name: AccountName) -> bool:
        """Return True if account is a debit account."""
        match self[account_name]:
            case Regular(t):
                return t in (T5.Asset, T5.Expense)
            case Contra(name):
                return not self.is_debit_account(name)
            case _:
                raise NotInChartError(account_name)

    def set(self, t: T5, account_name: str):
        """Add regular account."""
        self.data[account_name] = Regular(t)
        return self

    def offset(self, account_name: str, contra_account_name: str):
        """Add contra account."""
        if account_name not in self.keys():
            raise NotInChartError(account_name)
        self.data[contra_account_name] = Contra(account_name)
        return self

    def t_account_class(self, account_name: str) -> Type["TAccount"]:
        """Return T-account class constructor for a given account name."""
        return DebitAccount if self.is_debit_account(account_name) else CreditAccount

    def to_ledger(self) -> "Ledger":
        """Create ledger."""
        return Ledger(
            {
                account_name: self.t_account_class(account_name)()
                for account_name in self.keys()
            }
        )

    def opening_entry(self, opening_balances: dict) -> "Posting":
        """Create opening entry."""
        entry: Posting = []
        for account_name, amount in opening_balances.items():
            if self.is_debit_account(account_name):
                entry.append(Debit(account_name, amount))
            elif account_name in self.keys():
                entry.append(Credit(account_name, amount))
            else:
                raise NotInChartError(account_name)
        raise_if_not_balanced(entry)
        return entry

    def closing_pairs(self, earnings_account: AccountName) -> Iterator[Pair]:
        """Yield closing pairs for accounting period end.
        The closing pairs can point to сurrent earnings or retained earnings account.
        """
        for t in (T5.Income, T5.Expense):
            for account_name in self.by_type(t):
                for contra_name in self.find_contra_accounts(account_name):
                    yield contra_name, account_name
                yield account_name, earnings_account

    def by_type(self, t: T5) -> list[AccountName]:
        """Return account names for a given account type."""
        return [name for name, _t in self.items() if _t == Regular(t)]

    def find_contra_accounts(self, name: AccountName) -> list[AccountName]:
        """Find contra accounts for a given account name."""
        return [contra_name for contra_name, n in self.items() if n == Contra(name)]


@dataclass
class SingleEntry(ABC):
    """Base class for a single entry changes either a debit or a credit side of an account."""

    name: AccountName
    amount: Amount


class Debit(SingleEntry):
    """An entry that increases the debit side of an account."""


class Credit(SingleEntry):
    """An entry that increases the credit side of an account."""


"""Posting is a list of single entries that is assumed to be balanced."""
Posting = list[Debit | Credit]


def sum_of(posting: Posting, cls) -> Posting:
    """Return sum of debits or sum of credit entries."""
    return sum(entry.amount for entry in posting if isinstance(entry, cls))


def is_balanced(posting: Posting) -> bool:
    """Return True if posting is balanced."""
    return sum_of(posting, Debit) == sum_of(posting, Credit)


def raise_if_not_balanced(posting: Posting):
    """Raise error if posting is not balanced."""
    if not is_balanced(posting):
        raise AbacusError(f"Posting is not balanced: {posting}")


def double(debit: str, credit: str, amount: Amount) -> Posting:
    """Create double entry."""
    return [Debit(debit, amount), Credit(credit, amount)]


@dataclass
class TAccount(ABC):
    """Base class for T-account that holds amounts on the left and right sides.

    Parent class for:
      - DebitAccount
      - CreditAccount
    """

    left: Amount = Amount(0)
    right: Amount = Amount(0)

    def debit(self, amount: Amount):
        """Add amount to debit side of T-account."""
        self.left += amount

    def credit(self, amount: Amount):
        """Add amount to credit side of T-account."""
        self.right += amount

    def apply(self, entry: SingleEntry):
        """Apply single entry to T-account."""
        match entry:
            case Debit(_, amount):
                self.debit(amount)
            case Credit(_, amount):
                self.credit(amount)

    @abstractmethod
    def transfer(self, frm: AccountName, to: AccountName) -> Posting:
        """Transfer balance *frm* one account *to* another."""

    @property
    @abstractmethod
    def balance(self) -> Amount:
        """Return balance of T-account."""


class DebitAccount(TAccount):
    @property
    def balance(self) -> Amount:
        return self.left - self.right

    def transfer(self, frm: AccountName, to: AccountName):
        return double(to, frm, self.balance)


class CreditAccount(TAccount):
    @property
    def balance(self) -> Amount:
        return self.right - self.left

    def transfer(self, frm: AccountName, to: AccountName):
        return double(frm, to, self.balance)


class Ledger(UserDict[AccountName, TAccount]):
    def post_single(self, single_entry: SingleEntry) -> None:
        """Post single entry to ledger."""
        try:
            self.data[single_entry.name].apply(single_entry)
        except KeyError:
            raise NotInLedgerError(single_entry.name)

    def raise_if_not_found(self, posting: Posting) -> None:
        """Raise error if account name is not found."""
        for entry in posting:
            if entry.name not in self.keys():
                raise NotInLedgerError(entry.name)

    def post(self, entry: Posting) -> None:
        """Post entry to ledger."""
        raise_if_not_balanced(entry)
        self.raise_if_not_found(entry)
        for single_entry in entry:
            self.post_single(single_entry)

    def is_closed(self, chart_dict: ChartDict) -> bool:
        """Return True if ledger is has no temporary accounts."""
        for frm, _ in chart_dict.closing_pairs("_"):
            if frm in self.data.keys():
                return False
        return True

    @property
    def trial_balance(self):
        """Create trial balance from ledger."""

        def as_tuple(t_account):
            b = t_account.balance
            return (b, None) if isinstance(t_account, DebitAccount) else (None, b)

        return TrialBalance(
            {name: as_tuple(t_account) for name, t_account in self.items()}
        )

    # note: output similar to ReportDict, but .total() would be not meaningful
    @property
    def balances(self) -> dict[AccountName, Amount]:
        """Return account balances."""
        return {name: account.balance for name, account in self.items()}

    def net_balance(self, name: AccountName, contra_names: list[AccountName]) -> Amount:
        """Calculate net balance of an account by deducting the balances of its contra accounts."""
        contra_balances = sum(self[contra_name].balance for contra_name in contra_names)
        return self[name].balance - contra_balances

    def close_one(self, frm: AccountName, to: AccountName) -> Posting:
        """Close account and move its balances to another account."""
        entry = self.data[frm].transfer(frm, to)
        self.post(entry)
        del self.data[frm]
        return entry

    def close(self, closing_pairs: Sequence[Pair]) -> list[Posting]:
        """Close ledger at accounting period end."""
        return [self.close_one(frm, to) for frm, to in closing_pairs]

    def balance_sheet(self, chart_dict: ChartDict):
        """Create balance sheet from ledger."""
        fill = net_balances_factory(chart_dict, self)
        return BalanceSheet(
            assets=fill(T5.Asset),
            capital=fill(T5.Capital),
            liabilities=fill(T5.Liability),
        )

    def income_statement(self, chart_dict: ChartDict):
        """Create income statement from ledger."""
        fill = net_balances_factory(chart_dict, self)
        return IncomeStatement(income=fill(T5.Income), expenses=fill(T5.Expense))


class TrialBalance(UserDict[str, tuple[Amount, Amount]]):
    """Trial balance contains account names and balances."""


class ReportDict(UserDict[AccountName, Amount]):
    @property
    def total(self):
        return Amount(sum(self.data.values()))


def net_balances_factory(
    chart_dict: ChartDict, ledger: Ledger
) -> Callable[[T5], ReportDict]:
    """Create a function that calculates net balances based on chart and ledger."""

    def fill(t: T5) -> ReportDict:
        """Return net balances for a given account type."""
        result = ReportDict()
        for name in chart_dict.by_type(t):
            try:
                contra_accounts = chart_dict.find_contra_accounts(name)
                result[name] = ledger.net_balance(name, contra_accounts)
            except KeyError:  # protect from current earnings account not in ledger
                pass
        return result

    return fill


class Report:
    """Base class for financial reports."""


@dataclass
class IncomeStatement(Report):
    income: ReportDict
    expenses: ReportDict

    @property
    def net_earnings(self):
        """Calculate net earnings as income less expenses."""
        return self.income.total - self.expenses.total


@dataclass
class BalanceSheet(Report):
    assets: ReportDict
    capital: ReportDict
    liabilities: ReportDict

    def is_balanced(self) -> bool:
        """Return True if assets equal liabilities plus capital."""
        return self.assets.total == (self.capital.total + self.liabilities.total)
